#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2008 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Timers;
using HeuristicLab.Hive.Slave.Common;
using HeuristicLab.Hive.Slave.Communication;
using System.Diagnostics;
using HeuristicLab.Hive.Contracts.BusinessObjects;
using HeuristicLab.Hive.Contracts;
using HeuristicLab.Hive.Slave.Core.ConfigurationManager;
using HeuristicLab.Tracing;
using System.Threading;
using HeuristicLab.Hive.Slave.Communication.SlaveFacade;

namespace HeuristicLab.Hive.Slave.Core {
  /// <summary>
  /// Heartbeat class. It sends every x ms a heartbeat to the server and receives a Message
  /// </summary>
  public class HeartbeatManager {

    private bool offline;

    public TimeSpan Interval { get; set; }

    private Thread heartBeatThread;

    private static object locker = new object();

    public HeartbeatManager() {
      Interval = new TimeSpan(0,0,10);
    }

    public HeartbeatManager(TimeSpan interval) {
      Interval = interval;
    }

    private WcfService wcfService;

    private bool abortThreadPending;

    ReaderWriterLockSlim heartBeatThreadIsSleepingLock = new ReaderWriterLockSlim();
    private bool heartBeatThreadIsSleeping = false;

    /// <summary>
    /// Starts the Heartbeat signal.
    /// </summary>
    public void StartHeartbeat() {
      wcfService = WcfService.Instance;
      wcfService.ProcessHeartBeatCompleted += new EventHandler<ProcessHeartBeatCompletedEventArgs>(wcfService_ProcessHeartBeatCompleted);
      abortThreadPending = false;
      heartBeatThread = GetHeartBeatThread();
      heartBeatThread.Start();
    }

    private Thread GetHeartBeatThread() {
      return new Thread(() => {
        while (!abortThreadPending) {
          try {
            lock (locker) {
              if (wcfService.ConnState != NetworkEnum.WcfConnState.Connected) {
                wcfService.Connect(); // Login happens automatically upon successufl connection
              }
              if (!wcfService.LoggedIn) {
                wcfService.Login(ConfigManager.Instance.GetClientInfo()); // if login faild previously try again
              }

              if (wcfService.LoggedIn) {
                SlaveDto info = ConfigManager.Instance.GetClientInfo();

                HeartBeatData heartBeatData = new HeartBeatData {
                  SlaveId = info.Id,
                  FreeCores = info.NrOfCores - ConfigManager.Instance.GetUsedCores(),
                  FreeMemory = GetFreeMemory(),
                  JobProgress = ConfigManager.Instance.GetProgressOfAllJobs(),
                  IsAllowedToCalculate = UptimeManager.Instance.IsAllowedToCalculate() && UptimeManager.Instance.CalendarAvailable
                };

                if (!heartBeatData.IsAllowedToCalculate) {
                  // stop all running jobs and send snapshots to server
                  MessageQueue.GetInstance().AddMessage(MessageContainer.MessageType.UptimeLimitDisconnect);
                }

                Logger.Debug("Sending Heartbeat: " + heartBeatData);
                wcfService.ProcessHeartBeatSync(heartBeatData);                
              }

            } // lock
          }
          catch (Exception e) {
            Logger.Error("Heartbeat Thread failed badly: " + e.Message);
          }

          try {
            heartBeatThreadIsSleepingLock.EnterWriteLock();
            heartBeatThreadIsSleeping = true;
            heartBeatThreadIsSleepingLock.ExitWriteLock();

            Thread.Sleep(Interval);

            heartBeatThreadIsSleepingLock.EnterWriteLock();
            heartBeatThreadIsSleeping = false;
            heartBeatThreadIsSleepingLock.ExitWriteLock();
          }
          catch (ThreadInterruptedException e) {
            Logger.Debug("Heartbeat sleep interrupted");
          }
        } // while
        abortThreadPending = false;
        Logger.Debug("Heartbeat thread stopped");
      });
    }

    private int GetFreeMemory() {
      PerformanceCounter counter = new PerformanceCounter("Memory", "Available Bytes", true);
      int mb = (int)(counter.NextValue() / 1024 / 1024);
      return mb;
    }

    void wcfService_ProcessHeartBeatCompleted(object sender, ProcessHeartBeatCompletedEventArgs e) {
      Logger.Debug("Heartbeat Response received");
      e.Result.ActionRequest.ForEach(mc => MessageQueue.GetInstance().AddMessage(mc));
    }

    public void StopHeartBeat() {
      abortThreadPending = true;
      heartBeatThread.Interrupt();
      heartBeatThread.Join();
    }

    /// <summary>
    /// use this method to singalize there is work to do (to avoid the waiting period if its clear that actions are required)
    /// </summary>
    public void InterruptHeartBeatThread() {
      // deal with ReadWriteLock to avoid interrupting the thread while it is in SleepOrWait due to another reason (like waiting for a resource in a lock-statement)
      heartBeatThreadIsSleepingLock.EnterReadLock();
      Debug.WriteLine("InterruptHeartBeatThread");
      if (heartBeatThreadIsSleeping) {
        Debug.WriteLine("will interrupt heartbeatthread");
        heartBeatThread.Interrupt();
      } else {
        Debug.WriteLine("will not interrupt heartbeatthread");
      }
      heartBeatThreadIsSleepingLock.ExitReadLock();
    }
  }
}
